"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const stringHash = require("string-hash");
const Type_1 = require("./Type");
const TypeAttributes_1 = require("./TypeAttributes");
const Support_1 = require("./support/Support");
const Containers_1 = require("./support/Containers");
function debugStringForType(t) {
    const target = followTargetType(t);
    if (t === target) {
        return t.kind;
    }
    return `${t.kind} (${target.kind})`;
}
class Transformer {
    constructor(kind, sourceTypeRef) {
        this.kind = kind;
        this.sourceTypeRef = sourceTypeRef;
    }
    get sourceType() {
        return this.sourceTypeRef.deref()[0];
    }
    /** This must return a newly constructed set. */
    getChildren() {
        return new Set([this.sourceType]);
    }
    equals(other) {
        if (!(other instanceof Transformer))
            return false;
        return this.sourceTypeRef.equals(other.sourceTypeRef);
    }
    hashCode() {
        return this.sourceTypeRef.hashCode();
    }
    debugDescription() {
        return `${debugStringForType(this.sourceType)} -> ${this.kind}`;
    }
    debugPrintContinuations(_indent) {
        return;
    }
    debugPrint(indent) {
        console.log(Support_1.indentationString(indent) + this.debugDescription());
        this.debugPrintContinuations(indent + 1);
    }
}
exports.Transformer = Transformer;
class ProducerTransformer extends Transformer {
    constructor(kind, sourceTypeRef, consumer) {
        super(kind, sourceTypeRef);
        this.consumer = consumer;
    }
    getChildren() {
        const children = super.getChildren();
        if (this.consumer === undefined)
            return children;
        return Containers_1.setUnionInto(children, this.consumer.getChildren());
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof ProducerTransformer))
            return false;
        return Containers_1.areEqual(this.consumer, other.consumer);
    }
    hashCode() {
        const h = super.hashCode();
        return Support_1.addHashCode(h, Containers_1.hashCodeOf(this.consumer));
    }
    debugPrintContinuations(indent) {
        if (this.consumer === undefined)
            return;
        this.consumer.debugPrint(indent);
    }
}
exports.ProducerTransformer = ProducerTransformer;
class MatchTransformer extends Transformer {
    constructor(kind, sourceTypeRef, transformer) {
        super(kind, sourceTypeRef);
        this.transformer = transformer;
    }
    getChildren() {
        return Containers_1.setUnionInto(super.getChildren(), this.transformer.getChildren());
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof MatchTransformer))
            return false;
        return this.transformer.equals(other.transformer);
    }
    hashCode() {
        const h = super.hashCode();
        return Support_1.addHashCode(h, this.transformer.hashCode());
    }
    debugPrintContinuations(indent) {
        this.transformer.debugPrint(indent);
    }
}
exports.MatchTransformer = MatchTransformer;
class DecodingTransformer extends ProducerTransformer {
    constructor(sourceTypeRef, consumer) {
        super("decode", sourceTypeRef, consumer);
    }
    reverse(targetTypeRef, continuationTransformer) {
        if (continuationTransformer !== undefined) {
            return Support_1.panic("Reversing a decoding transformer cannot have a continuation");
        }
        if (this.consumer === undefined) {
            return new EncodingTransformer(targetTypeRef);
        }
        else {
            return this.consumer.reverse(targetTypeRef, new EncodingTransformer(this.consumer.sourceTypeRef));
        }
    }
    reconstitute(builder) {
        return new DecodingTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), Support_1.mapOptional(xfer => xfer.reconstitute(builder), this.consumer));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        return other instanceof StringProducerTransformer;
    }
}
exports.DecodingTransformer = DecodingTransformer;
class EncodingTransformer extends Transformer {
    constructor(sourceTypeRef) {
        super("encode", sourceTypeRef);
    }
    reverse(_targetTypeRef, _continuationTransformer) {
        return Support_1.panic("Can't reverse encoding transformer");
    }
    reconstitute(builder) {
        return new EncodingTransformer(builder.reconstituteTypeRef(this.sourceTypeRef));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof EncodingTransformer))
            return false;
        return true;
    }
}
exports.EncodingTransformer = EncodingTransformer;
class ChoiceTransformer extends Transformer {
    constructor(sourceTypeRef, transformers) {
        super("choice", sourceTypeRef);
        this.transformers = transformers;
    }
    getChildren() {
        let children = super.getChildren();
        for (const xfer of this.transformers) {
            Containers_1.setUnionInto(children, xfer.getChildren());
        }
        return children;
    }
    reverse(targetTypeRef, continuationTransformer) {
        const transformers = this.transformers.map(xfer => xfer.reverse(targetTypeRef, continuationTransformer));
        return new ChoiceTransformer(targetTypeRef, transformers);
    }
    reconstitute(builder) {
        return new ChoiceTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), this.transformers.map(xfer => xfer.reconstitute(builder)));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof ChoiceTransformer))
            return false;
        return Containers_1.areEqual(this.transformers, other.transformers);
    }
    hashCode() {
        const h = super.hashCode();
        return Support_1.addHashCode(h, Containers_1.hashCodeOf(this.transformers));
    }
    debugPrintContinuations(indent) {
        for (const xfer of this.transformers) {
            xfer.debugPrint(indent);
        }
    }
}
exports.ChoiceTransformer = ChoiceTransformer;
class DecodingChoiceTransformer extends Transformer {
    constructor(sourceTypeRef, nullTransformer, integerTransformer, doubleTransformer, boolTransformer, stringTransformer, arrayTransformer, objectTransformer) {
        super("decoding-choice", sourceTypeRef);
        this.nullTransformer = nullTransformer;
        this.integerTransformer = integerTransformer;
        this.doubleTransformer = doubleTransformer;
        this.boolTransformer = boolTransformer;
        this.stringTransformer = stringTransformer;
        this.arrayTransformer = arrayTransformer;
        this.objectTransformer = objectTransformer;
    }
    getChildren() {
        let children = super.getChildren();
        if (this.nullTransformer !== undefined) {
            Containers_1.setUnionInto(children, this.nullTransformer.getChildren());
        }
        if (this.integerTransformer !== undefined) {
            Containers_1.setUnionInto(children, this.integerTransformer.getChildren());
        }
        if (this.doubleTransformer !== undefined) {
            Containers_1.setUnionInto(children, this.doubleTransformer.getChildren());
        }
        if (this.boolTransformer !== undefined) {
            Containers_1.setUnionInto(children, this.boolTransformer.getChildren());
        }
        if (this.stringTransformer !== undefined) {
            Containers_1.setUnionInto(children, this.stringTransformer.getChildren());
        }
        if (this.arrayTransformer !== undefined) {
            Containers_1.setUnionInto(children, this.arrayTransformer.getChildren());
        }
        if (this.objectTransformer !== undefined) {
            Containers_1.setUnionInto(children, this.objectTransformer.getChildren());
        }
        return children;
    }
    reverse(targetTypeRef, continuationTransformer) {
        Support_1.assert(continuationTransformer === undefined, "Reversing a decoding transformer can't have a target transformer");
        let transformers = [];
        function addCase(transformer) {
            if (transformer === undefined)
                return;
            transformers.push(transformer.reverse(targetTypeRef, new EncodingTransformer(transformer.sourceTypeRef)));
        }
        addCase(this.nullTransformer);
        addCase(this.integerTransformer);
        addCase(this.doubleTransformer);
        addCase(this.boolTransformer);
        addCase(this.stringTransformer);
        addCase(this.arrayTransformer);
        addCase(this.objectTransformer);
        return new ChoiceTransformer(targetTypeRef, transformers);
    }
    reconstitute(builder) {
        function reconstitute(xf) {
            if (xf === undefined)
                return undefined;
            return xf.reconstitute(builder);
        }
        return new DecodingChoiceTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), reconstitute(this.nullTransformer), reconstitute(this.integerTransformer), reconstitute(this.doubleTransformer), reconstitute(this.boolTransformer), reconstitute(this.stringTransformer), reconstitute(this.arrayTransformer), reconstitute(this.objectTransformer));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof DecodingChoiceTransformer))
            return false;
        if (!Containers_1.areEqual(this.nullTransformer, other.nullTransformer))
            return false;
        if (!Containers_1.areEqual(this.integerTransformer, other.integerTransformer))
            return false;
        if (!Containers_1.areEqual(this.doubleTransformer, other.doubleTransformer))
            return false;
        if (!Containers_1.areEqual(this.boolTransformer, other.boolTransformer))
            return false;
        if (!Containers_1.areEqual(this.stringTransformer, other.stringTransformer))
            return false;
        if (!Containers_1.areEqual(this.arrayTransformer, other.arrayTransformer))
            return false;
        if (!Containers_1.areEqual(this.objectTransformer, other.objectTransformer))
            return false;
        return true;
    }
    hashCode() {
        let h = super.hashCode();
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.nullTransformer));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.integerTransformer));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.doubleTransformer));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.boolTransformer));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.stringTransformer));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.arrayTransformer));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.objectTransformer));
        return h;
    }
    debugPrintContinuations(indent) {
        if (this.nullTransformer !== undefined)
            this.nullTransformer.debugPrint(indent);
        if (this.integerTransformer !== undefined)
            this.integerTransformer.debugPrint(indent);
        if (this.doubleTransformer !== undefined)
            this.doubleTransformer.debugPrint(indent);
        if (this.boolTransformer !== undefined)
            this.boolTransformer.debugPrint(indent);
        if (this.stringTransformer !== undefined)
            this.stringTransformer.debugPrint(indent);
        if (this.arrayTransformer !== undefined)
            this.arrayTransformer.debugPrint(indent);
        if (this.objectTransformer !== undefined)
            this.objectTransformer.debugPrint(indent);
    }
}
exports.DecodingChoiceTransformer = DecodingChoiceTransformer;
class UnionMemberMatchTransformer extends MatchTransformer {
    constructor(sourceTypeRef, transformer, _memberTypeRef) {
        super("union-member-match", sourceTypeRef, transformer);
        this._memberTypeRef = _memberTypeRef;
    }
    get sourceType() {
        const t = this.sourceTypeRef.deref()[0];
        if (!(t instanceof Type_1.UnionType)) {
            return Support_1.panic("The source of a union member match transformer must be a union type");
        }
        return t;
    }
    get memberType() {
        return this._memberTypeRef.deref()[0];
    }
    getChildren() {
        return super.getChildren().add(this.memberType);
    }
    reverse(_targetTypeRef, _continuationTransformer) {
        return Support_1.panic("Can't reverse union member match transformer");
    }
    reconstitute(builder) {
        return new UnionMemberMatchTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), this.transformer.reconstitute(builder), builder.reconstituteTypeRef(this._memberTypeRef));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof UnionMemberMatchTransformer))
            return false;
        return this._memberTypeRef.equals(other._memberTypeRef);
    }
    hashCode() {
        const h = super.hashCode();
        return Support_1.addHashCode(h, this._memberTypeRef.hashCode());
    }
    debugDescription() {
        return `${super.debugDescription()} - member: ${debugStringForType(this.memberType)}`;
    }
}
exports.UnionMemberMatchTransformer = UnionMemberMatchTransformer;
/**
 * This matches strings and enum cases.
 */
class StringMatchTransformer extends MatchTransformer {
    constructor(sourceTypeRef, transformer, stringCase) {
        super("string-match", sourceTypeRef, transformer);
        this.stringCase = stringCase;
    }
    get sourceType() {
        const t = this.sourceTypeRef.deref()[0];
        if (!(t instanceof Type_1.EnumType) && !(t instanceof Type_1.PrimitiveType && t.kind === "string")) {
            return Support_1.panic("The source of a string match transformer must be an enum or string type");
        }
        return t;
    }
    reverse(targetTypeRef, continuationTransformer) {
        return this.transformer.reverse(targetTypeRef, new StringProducerTransformer(this.transformer.sourceTypeRef, continuationTransformer, this.stringCase));
    }
    reconstitute(builder) {
        return new StringMatchTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), this.transformer.reconstitute(builder), this.stringCase);
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof StringMatchTransformer))
            return false;
        return this.stringCase !== other.stringCase;
    }
    hashCode() {
        const h = super.hashCode();
        return Support_1.addHashCode(h, stringHash(this.stringCase));
    }
    debugDescription() {
        return `${super.debugDescription()} - case: ${this.stringCase}`;
    }
}
exports.StringMatchTransformer = StringMatchTransformer;
class UnionInstantiationTransformer extends Transformer {
    constructor(sourceTypeRef) {
        super("union-instantiation", sourceTypeRef);
    }
    reverse(targetTypeRef, continuationTransformer) {
        if (continuationTransformer === undefined) {
            return Support_1.panic("Union instantiation transformer reverse must have a continuation");
        }
        return new UnionMemberMatchTransformer(targetTypeRef, continuationTransformer, this.sourceTypeRef);
    }
    reconstitute(builder) {
        return new UnionInstantiationTransformer(builder.reconstituteTypeRef(this.sourceTypeRef));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        return other instanceof UnionInstantiationTransformer;
    }
}
exports.UnionInstantiationTransformer = UnionInstantiationTransformer;
/**
 * Produces a string or an enum case.
 */
class StringProducerTransformer extends ProducerTransformer {
    constructor(sourceTypeRef, consumer, result) {
        super("string-producer", sourceTypeRef, consumer);
        this.result = result;
    }
    reverse(targetTypeRef, continuationTransformer) {
        if (continuationTransformer === undefined) {
            return Support_1.panic("Reversing a string producer transformer must have a continuation");
        }
        if (this.consumer === undefined) {
            return new StringMatchTransformer(targetTypeRef, continuationTransformer, this.result);
        }
        else {
            return this.consumer.reverse(targetTypeRef, new StringMatchTransformer(this.consumer.sourceTypeRef, continuationTransformer, this.result));
        }
    }
    reconstitute(builder) {
        return new StringProducerTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), Support_1.mapOptional(xfer => xfer.reconstitute(builder), this.consumer), this.result);
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        if (!(other instanceof StringProducerTransformer))
            return false;
        return this.result === other.result;
    }
    hashCode() {
        const h = super.hashCode();
        return Support_1.addHashCode(h, Containers_1.hashCodeOf(this.consumer));
    }
    debugDescription() {
        return `${super.debugDescription()} - result: ${this.result}`;
    }
}
exports.StringProducerTransformer = StringProducerTransformer;
class ParseDateTimeTransformer extends ProducerTransformer {
    constructor(sourceTypeRef, consumer) {
        super("parse-date-time", sourceTypeRef, consumer);
    }
    reverse(targetTypeRef, continuationTransformer) {
        if (this.consumer === undefined) {
            return new StringifyDateTimeTransformer(targetTypeRef, continuationTransformer);
        }
        else {
            return this.consumer.reverse(targetTypeRef, new StringifyDateTimeTransformer(this.consumer.sourceTypeRef, continuationTransformer));
        }
    }
    reconstitute(builder) {
        return new ParseDateTimeTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), Support_1.mapOptional(xfer => xfer.reconstitute(builder), this.consumer));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        return other instanceof ParseDateTimeTransformer;
    }
}
exports.ParseDateTimeTransformer = ParseDateTimeTransformer;
class StringifyDateTimeTransformer extends ProducerTransformer {
    constructor(sourceTypeRef, consumer) {
        super("stringify-date-time", sourceTypeRef, consumer);
    }
    reverse(targetTypeRef, continuationTransformer) {
        if (this.consumer === undefined) {
            return new ParseDateTimeTransformer(targetTypeRef, continuationTransformer);
        }
        else {
            return this.consumer.reverse(targetTypeRef, new ParseDateTimeTransformer(this.consumer.sourceTypeRef, continuationTransformer));
        }
    }
    reconstitute(builder) {
        return new StringifyDateTimeTransformer(builder.reconstituteTypeRef(this.sourceTypeRef), Support_1.mapOptional(xfer => xfer.reconstitute(builder), this.consumer));
    }
    equals(other) {
        if (!super.equals(other))
            return false;
        return other instanceof StringifyDateTimeTransformer;
    }
}
exports.StringifyDateTimeTransformer = StringifyDateTimeTransformer;
class Transformation {
    constructor(_targetTypeRef, transformer) {
        this._targetTypeRef = _targetTypeRef;
        this.transformer = transformer;
    }
    get sourceType() {
        return this.transformer.sourceType;
    }
    get targetType() {
        return this._targetTypeRef.deref()[0];
    }
    get reverse() {
        return new Transformation(this.transformer.sourceTypeRef, this.transformer.reverse(this._targetTypeRef, undefined));
    }
    getChildren() {
        return this.transformer.getChildren().add(this.targetType);
    }
    reconstitute(builder) {
        return new Transformation(builder.reconstituteTypeRef(this._targetTypeRef), this.transformer.reconstitute(builder));
    }
    equals(other) {
        if (!(other instanceof Transformation))
            return false;
        return this._targetTypeRef.equals(other._targetTypeRef) && this.transformer.equals(other.transformer);
    }
    hashCode() {
        let h = this._targetTypeRef.hashCode();
        h = Support_1.addHashCode(h, this.transformer.hashCode());
        return h;
    }
    debugPrint() {
        this.transformer.debugPrint(0);
        console.log(`-> ${debugStringForType(this.targetType)}`);
    }
}
exports.Transformation = Transformation;
class TransformationTypeAttributeKind extends TypeAttributes_1.TypeAttributeKind {
    constructor() {
        super("transformation");
    }
    get inIdentity() {
        return true;
    }
    children(xf) {
        return xf.getChildren();
    }
    reconstitute(builder, xf) {
        return xf.reconstitute(builder);
    }
    stringify(_) {
        return "transformation";
    }
}
exports.transformationTypeAttributeKind = new TransformationTypeAttributeKind();
function transformationForType(t) {
    return exports.transformationTypeAttributeKind.tryGetInAttributes(t.getAttributes());
}
exports.transformationForType = transformationForType;
function followTargetType(t) {
    for (;;) {
        const xf = transformationForType(t);
        if (xf === undefined)
            return t;
        t = xf.targetType;
    }
}
exports.followTargetType = followTargetType;
