import { TypeRef } from "./TypeBuilder";
import { TypeReconstituter, BaseGraphRewriteBuilder } from "./GraphRewriting";
import { TypeNames } from "./TypeNames";
import { TypeAttributes } from "./TypeAttributes";
export declare type DateTimeTypeKind = "date" | "time" | "date-time";
export declare type PrimitiveStringTypeKind = "string" | DateTimeTypeKind;
export declare type PrimitiveTypeKind = "none" | "any" | "null" | "bool" | "integer" | "double" | PrimitiveStringTypeKind;
export declare type NamedTypeKind = "class" | "enum" | "union";
export declare type TypeKind = PrimitiveTypeKind | NamedTypeKind | "array" | "object" | "map" | "intersection";
export declare type ObjectTypeKind = "object" | "map" | "class";
export declare function isPrimitiveStringTypeKind(kind: TypeKind): kind is PrimitiveStringTypeKind;
export declare function isNumberTypeKind(kind: TypeKind): kind is "integer" | "double";
export declare function isPrimitiveTypeKind(kind: TypeKind): kind is PrimitiveTypeKind;
export declare class TypeIdentity {
    private readonly _kind;
    private readonly _components;
    private readonly _hashCode;
    constructor(_kind: TypeKind, _components: ReadonlyArray<any>);
    equals(other: any): boolean;
    hashCode(): number;
}
export declare type MaybeTypeIdentity = TypeIdentity | undefined;
export declare abstract class Type {
    readonly typeRef: TypeRef;
    readonly kind: TypeKind;
    constructor(typeRef: TypeRef, kind: TypeKind);
    abstract getNonAttributeChildren(): ReadonlySet<Type>;
    getChildren(): ReadonlySet<Type>;
    getAttributes(): TypeAttributes;
    readonly hasNames: boolean;
    getNames(): TypeNames;
    getCombinedName(): string;
    readonly abstract isNullable: boolean;
    abstract isPrimitive(): this is PrimitiveType;
    readonly abstract identity: MaybeTypeIdentity;
    abstract reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>, canonicalOrder: boolean): void;
    readonly debugPrintKind: string;
    equals(other: any): boolean;
    hashCode(): number;
    protected abstract structuralEqualityStep(other: Type, conflateNumbers: boolean, queue: (a: Type, b: Type) => boolean): boolean;
    structurallyCompatible(other: Type, conflateNumbers?: boolean): boolean;
    getParentTypes(): ReadonlySet<Type>;
    getAncestorsNotInSet(set: ReadonlySet<TypeRef>): ReadonlySet<Type>;
}
export declare function primitiveTypeIdentity(kind: PrimitiveTypeKind, attributes: TypeAttributes): MaybeTypeIdentity;
export declare class PrimitiveType extends Type {
    readonly kind: PrimitiveTypeKind;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    getNonAttributeChildren(): ReadonlySet<Type>;
    readonly identity: MaybeTypeIdentity;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(_other: Type, _conflateNumbers: boolean, _queue: (a: Type, b: Type) => boolean): boolean;
}
export declare function arrayTypeIdentity(attributes: TypeAttributes, itemsRef: TypeRef): MaybeTypeIdentity;
export declare class ArrayType extends Type {
    private _itemsRef;
    readonly kind: "array";
    constructor(typeRef: TypeRef, _itemsRef?: TypeRef | undefined);
    setItems(itemsRef: TypeRef): undefined;
    private getItemsRef();
    readonly items: Type;
    getNonAttributeChildren(): ReadonlySet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    readonly identity: MaybeTypeIdentity;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(other: ArrayType, _conflateNumbers: boolean, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class GenericClassProperty<T> {
    readonly typeData: T;
    readonly isOptional: boolean;
    constructor(typeData: T, isOptional: boolean);
    equals(other: any): boolean;
    hashCode(): number;
}
export declare class ClassProperty extends GenericClassProperty<TypeRef> {
    constructor(typeRef: TypeRef, isOptional: boolean);
    readonly typeRef: TypeRef;
    readonly type: Type;
}
export declare function classTypeIdentity(attributes: TypeAttributes, properties: ReadonlyMap<string, ClassProperty>): MaybeTypeIdentity;
export declare function mapTypeIdentify(attributes: TypeAttributes, additionalPropertiesRef: TypeRef | undefined): MaybeTypeIdentity;
export declare class ObjectType extends Type {
    readonly isFixed: boolean;
    private _properties;
    private _additionalPropertiesRef;
    readonly kind: ObjectTypeKind;
    constructor(typeRef: TypeRef, kind: ObjectTypeKind, isFixed: boolean, _properties: ReadonlyMap<string, ClassProperty> | undefined, _additionalPropertiesRef: TypeRef | undefined);
    setProperties(properties: ReadonlyMap<string, ClassProperty>, additionalPropertiesRef: TypeRef | undefined): void;
    getProperties(): ReadonlyMap<string, ClassProperty>;
    getSortedProperties(): ReadonlyMap<string, ClassProperty>;
    private getAdditionalPropertiesRef();
    getAdditionalProperties(): Type | undefined;
    getNonAttributeChildren(): ReadonlySet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    readonly identity: MaybeTypeIdentity;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>, canonicalOrder: boolean): void;
    protected structuralEqualityStep(other: ObjectType, _conflateNumbers: boolean, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class ClassType extends ObjectType {
    kind: "class";
    constructor(typeRef: TypeRef, isFixed: boolean, properties: ReadonlyMap<string, ClassProperty> | undefined);
}
export declare class MapType extends ObjectType {
    readonly kind: "map";
    constructor(typeRef: TypeRef, valuesRef: TypeRef | undefined);
    readonly values: Type;
}
export declare function enumTypeIdentity(attributes: TypeAttributes, cases: ReadonlySet<string>): MaybeTypeIdentity;
export declare class EnumType extends Type {
    readonly cases: ReadonlySet<string>;
    kind: "enum";
    constructor(typeRef: TypeRef, cases: ReadonlySet<string>);
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    readonly identity: MaybeTypeIdentity;
    getNonAttributeChildren(): ReadonlySet<Type>;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(other: EnumType, _conflateNumbers: boolean, _queue: (a: Type, b: Type) => void): boolean;
}
export declare function setOperationCasesEqual(typesA: Iterable<Type>, typesB: Iterable<Type>, conflateNumbers: boolean, membersEqual: (a: Type, b: Type) => boolean): boolean;
export declare function setOperationTypeIdentity(kind: TypeKind, attributes: TypeAttributes, memberRefs: ReadonlySet<TypeRef>): MaybeTypeIdentity;
export declare function unionTypeIdentity(attributes: TypeAttributes, memberRefs: ReadonlySet<TypeRef>): MaybeTypeIdentity;
export declare function intersectionTypeIdentity(attributes: TypeAttributes, memberRefs: ReadonlySet<TypeRef>): MaybeTypeIdentity;
export declare abstract class SetOperationType extends Type {
    private _memberRefs;
    constructor(typeRef: TypeRef, kind: TypeKind, _memberRefs?: ReadonlySet<TypeRef> | undefined);
    setMembers(memberRefs: ReadonlySet<TypeRef>): void;
    protected getMemberRefs(): ReadonlySet<TypeRef>;
    readonly members: ReadonlySet<Type>;
    readonly sortedMembers: ReadonlySet<Type>;
    getNonAttributeChildren(): ReadonlySet<Type>;
    isPrimitive(): this is PrimitiveType;
    readonly identity: MaybeTypeIdentity;
    protected reconstituteSetOperation<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>, canonicalOrder: boolean, getType: (members: ReadonlySet<TypeRef> | undefined) => void): void;
    protected structuralEqualityStep(other: SetOperationType, conflateNumbers: boolean, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class IntersectionType extends SetOperationType {
    kind: "intersection";
    constructor(typeRef: TypeRef, memberRefs?: ReadonlySet<TypeRef>);
    readonly isNullable: boolean;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>, canonicalOrder: boolean): void;
}
export declare class UnionType extends SetOperationType {
    kind: "union";
    constructor(typeRef: TypeRef, memberRefs?: ReadonlySet<TypeRef>);
    setMembers(memberRefs: ReadonlySet<TypeRef>): void;
    readonly stringTypeMembers: ReadonlySet<Type>;
    findMember(kind: TypeKind): Type | undefined;
    readonly isNullable: boolean;
    readonly isCanonical: boolean;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>, canonicalOrder: boolean): void;
}
