"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const CompressedJSON_1 = require("./CompressedJSON");
const Support_1 = require("../support/Support");
const UnionBuilder_1 = require("../UnionBuilder");
const DateTime_1 = require("../DateTime");
const Type_1 = require("../Type");
const TypeAttributes_1 = require("../TypeAttributes");
const StringTypes_1 = require("../StringTypes");
function forEachArrayInNestedValueArray(va, f) {
    if (va.length === 0) {
        return;
    }
    if (Array.isArray(va[0])) {
        for (const x of va) {
            forEachArrayInNestedValueArray(x, f);
        }
    }
    else {
        f(va);
    }
}
function forEachValueInNestedValueArray(va, f) {
    forEachArrayInNestedValueArray(va, a => {
        for (const x of a) {
            f(x);
        }
    });
}
class InferenceUnionBuilder extends UnionBuilder_1.UnionBuilder {
    constructor(typeBuilder, _typeInference, _cjson, _fixed) {
        super(typeBuilder);
        this._typeInference = _typeInference;
        this._cjson = _cjson;
        this._fixed = _fixed;
    }
    makeObject(objects, typeAttributes, forwardingRef) {
        return this._typeInference.inferClassType(this._cjson, typeAttributes, objects, this._fixed, forwardingRef);
    }
    makeArray(arrays, _typeAttributes, forwardingRef) {
        return this.typeBuilder.getArrayType(this._typeInference.inferType(this._cjson, new Map(), arrays, this._fixed, forwardingRef));
    }
}
function canBeEnumCase(s) {
    if (s.length === 0)
        return true; // FIXME: Do we really want this?
    return !DateTime_1.isDate(s) && !DateTime_1.isTime(s) && !DateTime_1.isDateTime(s);
}
class TypeInference {
    constructor(_typeBuilder, _inferEnums, _inferDates) {
        this._typeBuilder = _typeBuilder;
        this._inferEnums = _inferEnums;
        this._inferDates = _inferDates;
    }
    inferType(cjson, typeAttributes, valueArray, fixed, forwardingRef) {
        const accumulator = new UnionBuilder_1.UnionAccumulator(true);
        forEachValueInNestedValueArray(valueArray, value => {
            const t = CompressedJSON_1.valueTag(value);
            switch (t) {
                case CompressedJSON_1.Tag.Null:
                    accumulator.addNull(TypeAttributes_1.emptyTypeAttributes);
                    break;
                case CompressedJSON_1.Tag.False:
                case CompressedJSON_1.Tag.True:
                    accumulator.addBool(TypeAttributes_1.emptyTypeAttributes);
                    break;
                case CompressedJSON_1.Tag.Integer:
                    accumulator.addInteger(TypeAttributes_1.emptyTypeAttributes);
                    break;
                case CompressedJSON_1.Tag.Double:
                    accumulator.addDouble(TypeAttributes_1.emptyTypeAttributes);
                    break;
                case CompressedJSON_1.Tag.InternedString:
                    if (this._inferEnums) {
                        const s = cjson.getStringForValue(value);
                        if (canBeEnumCase(s)) {
                            accumulator.addStringCase(s, 1, TypeAttributes_1.emptyTypeAttributes);
                        }
                        else {
                            accumulator.addStringType("string", TypeAttributes_1.emptyTypeAttributes);
                        }
                    }
                    else {
                        accumulator.addStringType("string", TypeAttributes_1.emptyTypeAttributes);
                    }
                    break;
                case CompressedJSON_1.Tag.UninternedString:
                    accumulator.addStringType("string", TypeAttributes_1.emptyTypeAttributes);
                    break;
                case CompressedJSON_1.Tag.Object:
                    accumulator.addObject(cjson.getObjectForValue(value), TypeAttributes_1.emptyTypeAttributes);
                    break;
                case CompressedJSON_1.Tag.Array:
                    accumulator.addArray(cjson.getArrayForValue(value), TypeAttributes_1.emptyTypeAttributes);
                    break;
                case CompressedJSON_1.Tag.Date:
                    accumulator.addStringType("string", TypeAttributes_1.emptyTypeAttributes, this._inferDates ? StringTypes_1.StringTypes.date : StringTypes_1.StringTypes.unrestricted);
                    break;
                case CompressedJSON_1.Tag.Time:
                    accumulator.addStringType("string", TypeAttributes_1.emptyTypeAttributes, this._inferDates ? StringTypes_1.StringTypes.time : StringTypes_1.StringTypes.unrestricted);
                    break;
                case CompressedJSON_1.Tag.DateTime:
                    accumulator.addStringType("string", TypeAttributes_1.emptyTypeAttributes, this._inferDates ? StringTypes_1.StringTypes.dateTime : StringTypes_1.StringTypes.unrestricted);
                    break;
                default:
                    return Support_1.assertNever(t);
            }
        });
        const unionBuilder = new InferenceUnionBuilder(this._typeBuilder, this, cjson, fixed);
        return unionBuilder.buildUnion(accumulator, false, typeAttributes, forwardingRef);
    }
    inferClassType(cjson, typeAttributes, objects, fixed, forwardingRef) {
        const propertyNames = [];
        const propertyValues = {};
        forEachArrayInNestedValueArray(objects, arr => {
            for (let i = 0; i < arr.length; i += 2) {
                const key = cjson.getStringForValue(arr[i]);
                const value = arr[i + 1];
                if (!Object.prototype.hasOwnProperty.call(propertyValues, key)) {
                    propertyNames.push(key);
                    propertyValues[key] = [];
                }
                propertyValues[key].push(value);
            }
        });
        const properties = new Map();
        for (const key of propertyNames) {
            const values = propertyValues[key];
            const t = this.inferType(cjson, new Map(), values, false);
            const isOptional = values.length < objects.length;
            properties.set(key, new Type_1.ClassProperty(t, isOptional));
        }
        if (fixed) {
            return this._typeBuilder.getUniqueClassType(typeAttributes, true, properties, forwardingRef);
        }
        else {
            return this._typeBuilder.getClassType(typeAttributes, properties, forwardingRef);
        }
    }
}
exports.TypeInference = TypeInference;
